__program__ = "catastrophy"
__version__ = "0.1.0"
__authors__ = ", ".join(["Darcy Jones", "James Hane"])
__date__ = "31 September 2021"
__email__ = "darcy.a.jones@curtin.edu.au"
__short_blurb__ = (
    "Script to predict lifestyle of filamentous plant pathogens using "
    "carbohydrate-active enzymes (CAZymes)."
)

__license__ = f"""{__program__}-{__version__}

{__short_blurb__}

Copyright (C) {__date__},  {__authors__}

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
